from dataclasses import InitVar, dataclass, field

import numpy as np  # type: ignore

from survival_evaluation.types import NumericArrayLike


def to_array(array_like: NumericArrayLike, to_boolean: bool = False) -> np.array:
    array = np.asarray(array_like)
    shape = np.shape(array)
    if len(shape) > 1:
        raise ValueError(
            f"Input should be a 1-d array. Got a shape of {shape} instead."
        )
    if np.any(array < 0):
        raise ValueError("All event times must be greater than or equal to zero.")
    if to_boolean:
        check_indicators(array)
        return array.astype(bool)
    return array


def check_indicators(indicators: np.array) -> None:
    if not all(np.logical_or(indicators == 0, indicators == 1)):
        raise ValueError(
            "Event indicators must be 0 or 1 where 0 indicates censorship and 1 is an event."
        )


def validate_size(
    event_times: NumericArrayLike,
    event_indicators: NumericArrayLike,
    predictions: NumericArrayLike,
):
    same_size = (
        np.shape(event_times) == np.shape(event_indicators) == np.shape(predictions)
    )
    if not same_size:
        raise ValueError("All three inputs must be of the same shape.")


@dataclass
class KaplanMeier:
    event_times: InitVar[np.array]
    event_indicators: InitVar[np.array]
    survival_times: np.array = field(init=False)
    survival_probabilities: np.array = field(init=False)

    def __post_init__(self, event_times, event_indicators):
        index = np.lexsort((event_indicators, event_times))
        unique_times = np.unique(event_times[index], return_counts=True)
        self.survival_times = unique_times[0]
        population_count = np.flip(np.flip(unique_times[1]).cumsum())

        event_counter = np.append(0, unique_times[1].cumsum()[:-1])
        event_ind = list()
        for i in range(np.size(event_counter[:-1])):
            event_ind.append(event_counter[i])
            event_ind.append(event_counter[i + 1])
        event_ind.append(event_counter[-1])
        event_ind.append(len(event_indicators))
        events = np.add.reduceat(np.append(event_indicators[index], 0), event_ind)[::2]

        self.survival_probabilities = np.empty(population_count.size)
        survival_probability = 1
        counter = 0
        for population, event_num in zip(population_count, events):
            survival_probability *= 1 - event_num / population
            self.survival_probabilities[counter] = survival_probability
            counter += 1

    def predict(self, prediction_times: np.array):
        probability_index = np.digitize(prediction_times, self.survival_times)
        probability_index = np.where(
            probability_index == self.survival_times.size + 1,
            probability_index - 1,
            probability_index,
        )
        probabilities = np.append(1, self.survival_probabilities)[probability_index]

        return probabilities


@dataclass
class KaplanMeierArea(KaplanMeier):
    area_times: np.array = field(init=False)
    area_probabilities: np.array = field(init=False)
    area: np.array = field(init=False)

    def __post_init__(self, event_times, event_indicators):
        super().__post_init__(event_times, event_indicators)
        area_probabilities = np.append(1, self.survival_probabilities)
        area_times = np.append(0, self.survival_times)
        if self.survival_probabilities[-1] != 0:
            slope = (area_probabilities[-1] - 1) / area_times[-1]
            zero_survival = -1 / slope
            area_times = np.append(area_times, zero_survival)
            area_probabilities = np.append(area_probabilities, 0)

        area_diff = np.diff(area_times, 1)
        area = np.flip(np.flip(area_diff * area_probabilities[0:-1]).cumsum())

        self.area_times = np.append(area_times, np.inf)
        self.area_probabilities = area_probabilities
        self.area = np.append(area, 0)

    def best_guess(self, censor_times: np.array):
        surv_prob = self.predict(censor_times)
        censor_indexes = np.digitize(censor_times, self.area_times)
        censor_indexes = np.where(
            censor_indexes == self.area_times.size + 1,
            censor_indexes - 1,
            censor_indexes,
        )
        censor_area = (
            self.area_times[censor_indexes] - censor_times
        ) * self.area_probabilities[censor_indexes - 1]
        censor_area += self.area[censor_indexes]
        return censor_times + censor_area / surv_prob



from typing import Optional

import numpy as np  # type: ignore
from scipy.stats import chi2  # type: ignore

from survival_evaluation.types import NumericArrayLike
from survival_evaluation.utility import (
    KaplanMeier,
    KaplanMeierArea,
    to_array,
    validate_size,
)


# pylint: disable=too-many-arguments
def l1(
    event_times: NumericArrayLike,
    event_indicators: NumericArrayLike,
    predictions: NumericArrayLike,
    training_event_times: Optional[NumericArrayLike] = None,
    training_event_indicators: Optional[NumericArrayLike] = None,
    l1_type: str = "hinge",
) -> float:

    event_times = to_array(event_times)
    event_indicators = to_array(event_indicators, to_boolean=True)
    predictions = to_array(predictions)

    validate_size(event_times, event_indicators, predictions)
    if l1_type == "hinge":
        scores = event_times - predictions
        scores[~event_indicators] = np.maximum(scores[~event_indicators], 0)
        return np.mean(np.abs(scores))

    if l1_type == "margin":
        if training_event_times is None or training_event_indicators is None:
            raise ValueError(
                "If 'margin' is chosen, training set values must be included."
            )

        training_event_times = to_array(training_event_times)
        training_event_indicators = to_array(training_event_indicators, to_boolean=True)

        km_model = KaplanMeierArea(training_event_times, training_event_indicators)
        censor_times = event_times[~event_indicators]
        weights = 1 - km_model.predict(censor_times)
        best_guesses = km_model.best_guess(censor_times)

        scores = np.empty(predictions.size)
        scores[event_indicators] = (
            event_times[event_indicators] - predictions[event_indicators]
        )
        scores[~event_indicators] = weights * (
            best_guesses - predictions[~event_indicators]
        )
        weighted_multiplier = 1 / (np.sum(event_indicators) + np.sum(weights))
        return weighted_multiplier * np.sum(np.abs(scores))

    raise ValueError("L1 type must be either 'hinge' or 'margin'.")


# pylint: disable=too-many-arguments
def one_calibration(
    event_times: NumericArrayLike,
    event_indicators: NumericArrayLike,
    predictions: NumericArrayLike,
    time: float,
    bins: int = 10,
) -> float:

    event_times = to_array(event_times)
    event_indicators = to_array(event_indicators, to_boolean=True)
    predictions = 1 - to_array(predictions)

    prediction_order = np.argsort(-predictions)
    predictions = predictions[prediction_order]
    event_times = event_times[prediction_order]
    event_indicators = event_indicators[prediction_order]

    # Can't do np.mean since split array may be of different sizes.
    binned_event_times = np.array_split(event_times, bins)
    binned_event_indicators = np.array_split(event_indicators, bins)
    probability_means = [np.mean(x) for x in np.array_split(predictions, bins)]
    hosmer_lemeshow = 0
    for b in range(bins):
        prob = probability_means[b]
        if prob == 1.0:
            raise ValueError(
                "One-Calibration is not well defined: the risk"
                f"probability of the {b}th bin was {prob}."
            )
        km_model = KaplanMeier(binned_event_times[b], binned_event_indicators[b])
        event_probability = 1 - km_model.predict(time)
        bin_count = len(binned_event_times[b])
        hosmer_lemeshow += (bin_count * event_probability - bin_count * prob) ** 2 / (
            bin_count * prob * (1 - prob)
        )

    return 1 - chi2.cdf(hosmer_lemeshow, bins - 1)


def d_calibration(
    event_indicators: NumericArrayLike,
    predictions: NumericArrayLike,
    bins: int = 10,
) -> float:

    event_indicators = to_array(event_indicators, to_boolean=True)
    predictions = to_array(predictions)

    # include minimum to catch if probability = 1.
    bin_index = np.minimum(np.floor(predictions * bins), bins - 1).astype(int)
    censored_bin_indexes = bin_index[~event_indicators]
    uncensored_bin_indexes = bin_index[event_indicators]

    censored_predictions = predictions[~event_indicators]
    censored_contribution = 1 - (censored_bin_indexes / bins) * (
        1 / censored_predictions
    )
    censored_following_contribution = 1 / (bins * censored_predictions)

    contribution_pattern = np.tril(np.ones([bins, bins]), k=-1).astype(bool)

    following_contributions = np.matmul(
        censored_following_contribution, contribution_pattern[censored_bin_indexes]
    )
    single_contributions = np.matmul(
        censored_contribution, np.eye(bins)[censored_bin_indexes]
    )
    uncensored_contributions = np.sum(np.eye(bins)[uncensored_bin_indexes], axis=0)
    bin_count = (
        single_contributions + following_contributions + uncensored_contributions
    )
    chi2_statistic = np.sum(
        np.square(bin_count - len(predictions) / bins) / (len(predictions) / bins)
    )
    return 1 - chi2.cdf(chi2_statistic, bins - 1)
